<?php if (!defined('PmWiki')) exit();

/*

 Username Authentification

 -- Copyright --

 Copyright 2004, 2005 by James McDuffie (ttq3sxl02@sneakemail.com)

 -- License --

 GNU GPL

 -- Description --

 This PmWiki plugin makes it possible to restrict access to the entire
 wiki by requring a username and password for certain actions. Individual
 pages and groups can still be password restricted in the same way that
 PmWiki does by default. However, user authentification happens before
 wiki password authentification. Additionally, the username doubles as
 the author name.

 The UserAuth package also contains two optional scripts that add 
 functionality. One of the scripts, userauth-pwchange.php adds the ability
 for users to change their own passwords. Another script, 
 userauth-admintool.php adds the ability for the admin user to edit
 user abilities and passwords as well as add new users. Please see the
 details within the scripts on how to install and configure them.

 By default this plugin reads username, password and ability information
 from a .htpasswd file. This file is created when the script is first run
 in the PmWiki "local" directory with the default restrictions that non 
 authorized users can only read pages. The htpasswd file has an additional
 third field where abilities are stored. This field should not interfer if
 you wish to use the same htpassword file with Apache basic authentification.
 A typical htpassword file would look like the following:

   GuestUser::read,rss
   LoggedInUser::pwchange
   FredFlintstone:HWxwafxAweaXefafdae:read_all,edit_all
   BarneyRubble:AsafasfAfadfadfad:read_group-PrivateArea,edit_group-PrivateArea,edit
   WilmaFlintstone:GFSDGfgdfgsdfgdf:edit,upload
   Admin:XwasgeGGGaaceexeadeE:admin

 The abilities field is a list of actions the user can perform separated by
 commas. Abilities correspond to the same type of actions defined by PmWiki in 
 the $DefaultPassword array as well as some custom actions:

   PmWiki Actions:
     admin, attr, edit, read, upload
       - If a PmWiki action is defined for a user then the user can perform the action
         for all groups that do not have a action_group-GroupName style ability defined.
   UserAuth Actions:
     admin
       - Admin is also checked explicitly by this plugin. This allows
         an admin user to be able to access any PmWiki password
         protected page without supplying the password.
     read_all, edit_all, attr_all, upload_all
       - Allows the user to perform action on any page, even if it is PmWiki
         password
         protected or protected by a group specific ability.
     read_group-GroupName, edit_group-GroupName, 
     attr_group-GroupName, upload_group-GroupName
       - Allows the user to perform action on all pages in GroupName, 
         for example:
         read_group-PrivateArea
         edit_group-PrivateArea
     read_page-PageName, edit_page-PageName, 
     attr_page-PageName, upload_page-PageName
       - Allows user to perform action on a specific page. This ability beats
         out any group level ability that is defined.
     
   
 -- Installation Instructions --

 * Unpack the distribution file into your PmWiki cookbook directory

 * Include the plugin from your config.php file:

   require_once("cookbook/userauth.php");

 * Load your PmWiki once to create the default .htpasswd file

 * Modify the created .htpasswd file to include a user that has "edit" ability.
   Use your wiki to create the password by appending ?action=crypt to the url
   to load the PmWiki password encryption page. Create the user's line in the
   htpasswd file in the usual Apace htpasswd format but make sure to add and
   extra field to contain the user's abilites. See the description above for
   an example of how to format the file.

 * Create the page Main.LoginPage so that the login page will show up within
   your wiki layout, otherwise a default standalone page will be used. Simply
   put the following on Main.LoginPage:

   [loginform]

   Main.LoginPage is just the default and the login page can be changed to
   whatever you like. See the configuration details below. 

 * Optionally install userauth-pwchange.php and userauth-admintool.php
   scripts. See scripts source for details.

 -- History --

 0.1 - Initial version
 0.2 - Added ability for allowing a user to view an entire view password 
       protected group
 0.3 - Updated for PmWiki v2. (BenWilson)
 0.4 - Changed loginform directive so that urls are not formatted by PmWiki
       Implemented group specific abilities that are needed whenever a group 
         specific abilities is defined for any user
       Improved greatly how forms and messages are displayed. Got rid of 
         Standalone form fall back
 0.5 - Added $IsUserLoggedIn variable for use by conditional markup
       Added CSS class tags userauthinput and userauthbutton for 
         controlling the look of input field and buttons
       Added $LoggedInUsername defintion for future expansion of actions that
         have abilities defined globally for all logged in users.
 0.6 - Now using BasicAuth from PmWiki instead of checking Wiki password directly
       First check user password as Wiki password in case user used the same 
         password to protect a page
       Renamed some variables and functions to make purpose even clearer
       Fixed issue with &nbsp in form instead of &nbsp;
       Implemented redirection configuration variables: $RedirectToPrevious 
         and $DefaultRedirectionPage
       Added a new class type, UserInstanceVars and implementation, 
         UserSessionVars to manage handling of instance username and 
         password information. This change will hopefully allow people to
         write code to store user info in cookies or whatever is needed for
         expanding functionality.
       Added page level abilities that rule over group level abilities just
         as a PmWiki page password would rule over a PmWiki group password.
       Added $[internationalization] substitutions to GetUserLoginForm() and
         GetWikiPasswordForm().
       Fixed infinite redirection problem that occurs when $LoginPage is 
         defined yet there is no read abilities defined for the GuestUser
       Changed markup for loginform from [loginform] to (:loginform:) to
         make syntax more consistent with PmWiki version 2's markup
       Fixed problem where the IsUserLoggedIn variable was always set to true
 0.61 - Fixed problem where previous page was not set when not using an
          embedded login page instead of a Wiki page with the loginform
          markup on it
        Fixed problem where CheckUserAuthPassword might display a prompt if
          called from another module. Now its safe to call UserAuth directly
          from another module as long as $authprompt = false
 0.62 - Fixed issue where "_all" attributes were not properly overriding
          wiki passwords as they should

 -- Configuration --

 The following variables can be defined before userauth.php is included in
 in config.php:

 $GuestUsername - Specifies the username that is used by default when no
 user is logged in.
   
 $LoginPage - Name of the page which has the "(:loginform:)" markup
 on it to display the login form.

 $LackProperAbilitiesFmt - Message displayed when the user unsuccesfully
 tries to login.

 $WrongUserOrPassFmt - Message displayed above login prompt when a user
 uses an incorrect username or password.

 $UserInfoObj - Specified the object used to obtain information on users
 and their abilities. By default HtPasswd.php will be used to grab user
 information unless the $UserInfoObj is already set. This allows someone 
 to write a new user info object that gets its information from a database
 or some other format. This new class would only need to implement the
 contract as can bee seen in the HtPasswd.php file. Only implement those
 methods below the line that says something to the effect in that file.

 $HtPasswdFile - Alternative location of the HtPassword file. Not used if
 $UserInfoObj is set to a different object before the script is included
 in config.php.

 $UserInstanceVars - Defines the class that manages user information such
 as username and password for the instance. This information is for the 
 current logged in user, or blank if the person accessing the site is not
 logged in. The default class, UserSessionVars will be used unless this
 variable is defined before UserAuth is loaded. If someone wants to handle
 instance user information with a method other than the $_SESSION array then
 create a class that has the same contract as UserSessionVars and set this
 variable.

 $RedirectToPrevious - This variable determines if upon succesful login
 or logout UserAuth redirects to the page that originally called the
 login or logout action. The default is true. If for instance you want
 to always redirect to $DefaultRedirectionPage then set this variable 
 to false.

 $DefaultRedirectionPage - If there is no information about which page
 orignially called the login or logout action then redirect to this page. 
 The default is to redirect to Main.HomePage when there is no previous page 
 information. One could possibly write a script that gets called before
 UserAuth and sets up this variable based upon things such as the
 current group, username or whatever.

 --- HTML Customization ---

 * The HTML that is displayed on the LoginPage can be easily changed. By 
 default the [loginform] inline replacement calls a function that returns 
 the login page. To specify different HTML then either redefine the
 definition of [loginform] in the Markup() or use some other method that
 correctly calls the login action.

 * If you wish to change your template to include a link that 
 displays "Login" or "Logout" depending on whether or not a user is 
 logged in then add the following to the template:

   <a href='$PageUrl?action=$LoginAction'>$LoginName</a>

 * The $IsUserLoggedIn variable can be use with the (:if :) conditional
 directive. To use it you need to add something like the following to
 you config.php:

   $Conditions['loggedin'] = "\$GLOBALS['IsUserLoggedIn']==\$condparm";
 
 Then you can use the following markup to display certain page data only
 for logged in users:
   
   (:if loggedin true:) [ Edit ] [ View ] (:if:)

 * The .userauthinput and userauthbutton CSS classes can be used to 
 control the look of input fields and buttons defined in the userauth class

*/

define(USER_AUTH_VERSION, '0.62'); 

// First we need to load the default class that handles username and 
// passwords for the instance
if(!isset($UserInstanceVars)) {
  require_once("userauth/UserSessionVars.php");
  $UserInstanceVars = new UserSessionVars();
}

// Now see if username and password information were passed through
// a post or get call. Store that information in the instance class if
// it is available
if (@$_REQUEST['username']) 
  $UserInstanceVars->SetInstanceUsername($_REQUEST['username']);
if (@$_REQUEST['user_pw'])  
  $UserInstanceVars->SetInstancePassword($_REQUEST['user_pw']);

// Override PmWiki variables
$Author = $UserInstanceVars->GetInstanceUsername();
$AuthFunction = "UserAuth";

// This variable will be set to true once the user has been verified
$IsUserLoggedIn = false;

// Configuration that your probably do not want to change
SDV($HandleActions['logout'], 'HandleLogout');
SDV($HandleActions['login'], 'HandleLogin');

SDV($LoginAction, GetLoginAction());
SDV($LoginName, ucfirst(GetLoginAction()));

// Customizable configuration variable
SDV($GuestUsername, "GuestUser");
SDV($LoggedInUsername, "LoggedInUser");

SDV($LoginPage, "Main.LoginPage");

SDV($LackProperAbilitiesFmt, "Insufficient privileges to perform action.");
SDV($WrongUserOrPassFmt, "Wrong username or password supplied.");

SDV($RedirectToPrevious, true);
SDV($DefaultRedirectionPage, "Main.HomePage");

Markup("loginform", "_end", "/\\(:loginform:\\)/", GetUserLoginForm(true));

# HtPasswd will be used unless the $UserInfoObj is already set.
# This allows someone to write a new user info object that gets
# its information from a database or some other format. This new
# class would only need to implement the same contract that
# HtPasswd.php does
if(!isset($UserInfoObj)) {
  require_once("userauth/HtPasswd.php");

  SDV($HtPasswdFile, getcwd() . "/local/.htpasswd");

  # If htpassword file does not exist then create a default empty one
  if (!file_exists($HtPasswdFile)) {
    $fp = @fopen($HtPasswdFile, "w");
    if ($fp) {
      fwrite($fp, "$GuestUsername::read\n");
      fclose($fp);
    }
  }

  // Object which contains username, password and abilities info
  $UserInfoObj = new HtPasswd($HtPasswdFile);

}


function UserAuth($pagename, $level, $authprompt=true) {
  global $UserInstanceVars, $UserInfoObj; 
  global $GuestUsername, $HtPasswdFile, $IsUserLoggedIn;

  $page = ReadPage($pagename);
  if (!$page) { return false; }

  // If no username has been specified then use a default name
  $auth_username = $UserInstanceVars->GetInstanceUsername();
  if ($auth_username=="") { $auth_username = $GuestUsername; }

  $userauth_check = 
    CheckUserAuthPassword( $auth_username, $UserInstanceVars->GetInstancePassword(), 
			   $level, $pagename, $authprompt );

  // If user auth has succeeded, we try to see if the user auth password is the
  // same as the pmwiki auth one, so we don't have to ask it twice
  $userpw_wiki_check = false;
  if ($userauth_check) {
    $userpw_wiki_check =
      CheckPmWikiPassword( $UserInstanceVars->GetInstancePassword(), 
			   $level, $pagename );
  }

  // If the user auth password doesn't work for pmwiki auth, then we try the
  // pmwiki auth password
  //
  // $_REQUEST is used since PmWiki's BasicAuth manages storing of Wiki passwords
  $wikipw_wiki_check = false;
  if (!$userpw_wiki_check) {
    $wikipw_wiki_check =
      CheckPmWikiPassword( $_REQUEST['wiki_pw'], $level, $pagename );
  }

  // Admin overrides everything even if a password is already specified
  //
  // "all" abilities only override their particular level type, they are
  // pseudo admin abilities.
  // 
  // Otherwise a userauth and pmwiki authentification check must succeed
  if ( ($userauth_check && $UserInfoObj->UserHasAbility($auth_username, 'admin')) ||
       ($userauth_check && $UserInfoObj->UserHasAbility($auth_username, $level . '_all')) ||
       ($userauth_check && ($userpw_wiki_check || $wikipw_wiki_check)) ) {

    // Set variable that can be used for conditional markup
    if($UserInstanceVars->GetInstanceUsername() != ''){
       $IsUserLoggedIn = true;
    }

    return $page;
  }

  // Return if we are not supposed to display any prompts, 
  // ie if called from another module
  if (!$authprompt) return false;

  if ( !$userauth_check ) {
    // Remove wrong login information from session
    $UserInstanceVars->ClearInstanceVariables();

    HandleLogin( $pagename );
  } elseif ( !$wiki_check ) {
    PrintEmbeddedPage( $pagename, GetWikiPasswordForm($pagename) );
  }
 
  exit;
}

function CheckUserAuthPassword($auth_username, $auth_password, $level, $pagename, $authprompt) {
  global $GuestUsername, $LoggedInUsername, $UserInfoObj;
  global $LackProperAbilitiesFmt, $WrongUserOrPassFmt;

  // Access username from htpasswd file and see if level and password match
  $stored_password = $UserInfoObj->GetUserPassword($auth_username);

  // Retrieve group name for group specific permissions
  $name  = FmtPageName('$Name',$pagename);
  $group = FmtPageName('$Group',$pagename);

  // Normally the userauth level to check the UserInfoObj is the same
  // as specified by PmWiki to UserAuth
  $userauth_level = $level;

  // Change checked level if group specific abilities exist user info object
  $group_specific_level = $level . "_group-" . $group;
  if($UserInfoObj->AnyUserHasAbility($group_specific_level)) {
    $userauth_level = $group_specific_level;
  }

  // Change checked level if page specific abilities exist user info object
  // page level abilities win over group level abilities
  $page_specific_level = $level . "_page-" . $name;
  if($UserInfoObj->AnyUserHasAbility($page_specific_level)) {
    $userauth_level = $page_specific_level;
  }

  // Only allow an empty password for the guest user
  if ( $auth_username == $GuestUsername && 
       $stored_password == "" && $auth_password == "" ) {

    if ($UserInfoObj->UserHasAbility($GuestUsername, $userauth_level)) return true;

  } elseif ( crypt($auth_password, $stored_password) == $stored_password ) {

    if ($UserInfoObj->UserHasAbility($auth_username, 'admin')) return true;
    if ($UserInfoObj->UserHasAbility($GuestUsername, $userauth_level)) return true;
    if ($UserInfoObj->UserHasAbility($LoggedInUsername, $userauth_level)) return true;
    if ($UserInfoObj->UserHasAbility($auth_username, $userauth_level)) return true;
    if ($UserInfoObj->UserHasAbility($auth_username, $level . '_all')) return true;

    // If we are not supposed to display prompts then just return false
    if(!$authprompt) return false;

    // If the above failed then we need to tell the user the action is not
    // possible, but not because of a bad password
    PrintEmbeddedPage( $pagename, $LackProperAbilitiesFmt );
    exit;

  } else {
    $_SESSION['auth_message'] = $WrongUserOrPassFmt;
  }

  return false;
}



function CheckPmWikiPassword($page_passwd, $level, $pagename) {
  global $UserInfoObj;

  // Setup Wiki password we are testing as if it came from a post so PmWiki's 
  // BasicAuth can check it for us
  $_POST['authpw'] = $page_passwd;
 
  // Test wiki password using PmWiki's builtin BasicAuth
  return BasicAuth($pagename, $level, false);
}


function PrintEmbeddedPage($pagename, $prompt) {
  global $PageStartFmt, $PageEndFmt, $AuthFunction;

  // Temporarily disable authentification or else its possible
  // that PrintFmt could go into an infinite recursive calls of UserAuth
  // when the user does not have any permissions at all
  // The worst that can happen is a user will see any pages included
  // by the template.
  $OrigAuthFunction = $AuthFunction;
  $AuthFunction = "true";

  $page = ReadPage($pagename);
  PCache($pagename,$page);

  $AuthEmbeddedFmt = array( &$PageStartFmt,
			    $prompt,
			    &$PageEndFmt);

  PrintFmt($pagename, $AuthEmbeddedFmt);

  // Go back to the original auth function just in case there are any
  // calls after this function
  $AuthFunction = $OrigAuthFunction;
}

function PrintStandalonePage($prompt) {
  global $HTMLStartFmt, $HTMLEndFmt;

  $AuthEmbeddedFmt = array( &$HTMLStartFmt,
			    $prompt,
			    &$HTMLEndFmt);

  PrintFmt($pagename, $AuthEmbeddedFmt);
}

function GetUserLoginForm($no_title) {
  global $UserInstanceVars, $Author, $WikiTitle, $pagename;

  if( $UserInstanceVars->GetInstanceUsername() == "" ) {

    if (!$no_title) {
      $login_form .= "<h1>$[Login to] $WikiTitle</h1>";
    }
    
    if ($_SESSION['auth_message']) {
      $login_form .= $_SESSION['auth_message'] . "<P>\n";
      $_SESSION['auth_message'] = '';
    }

    $login_form .=
      "<form name='authform' action='{$_SERVER['REQUEST_URI']}?action=login' method='post'>
       <table>

       <tr>
           <td>$[Username]:</td>
           <td><input name='username' class='userauthinput' value='$author' /></td>
       </tr>

       <tr>
           <td>$[Password]:</td>
           <td><input name='user_pw' class='userauthinput' type='password' value='' /></td>
       </tr>
       <tr>
           <td align=left><input class='userauthbutton' type='submit' value='$[Login]' /></td>
           <td>&nbsp;</td> 
       </tr>
       </table>
       </form>";

  } else {
    $login_form = 
      "$[Logged in as]: " . $Author . "<br>
       <a href=\"" . $_SERVER['REQUEST_URI'] . "?action=logout\">$[Logout]</a>";
    
  }

  return FmtPageName($login_form, $pagename);
}

function GetWikiPasswordForm($pagename) {
  global $pagename;

  $password_form .=
    "<h1>$[Enter password for] $pagename</h1>
     <form name='authform' action='{$_SERVER['REQUEST_URI']}' method='post'>
     <table>

     <tr>
         <td>$[Password]:</td>
         <td><input name='wiki_pw' class='userauthinput' type='password'  value='' /></td>
     </tr>
     <tr>
         <td align=left><input class='userauthbutton' type='submit' value='$[Submit]' /></td>
         <td>&nbsp;</td> 
     </tr>
     </table>
     </form>";

  return FmtPageName($password_form, $pagename);
}

function GetLoginAction() {
  global $UserInstanceVars;

  if( $UserInstanceVars->GetInstanceUsername() == "" ) {
    return "login";
  } else {
    return "logout";
  }

}

function HandleLogin($pagename) {
  global $UserInstanceVars, $UserInfoObj, $GuestUsername, $LoginPage;
  global $RedirectToPrevious, $DefaultRedirectionPage;

  if( $UserInstanceVars->GetInstanceUsername() != "" ) {
    // Only redirect to the page that called the login method originally
    // if the $RedirectToPrevious configuration variable is set
    if ( $RedirectToPrevious && isset($_SESSION['previous_page'])) {
      $prev_page = $_SESSION['previous_page'];
      unset($_SESSION['previous_page']);

      // Redirect to a URL if we have a referrer else go to
      // the page name we are handling login for
      if (preg_match("/^http/i", $prev_page)) {
	RedirectURL($prev_page);
      } else {
	Redirect($prev_page);
      }

    } else {
      // If a user defined default redirection is set then redirect to that
      // otherwise redirect to whatever pagename was passed to this method
      if(isset($DefaultRedirectionPage)) {
	Redirect($DefaultRedirectionPage);
      } else {
	Redirect($pagename);
      }
    }
  } else {
    // Remember the page that called for login, either by 
    // tracking the referer or by going to the page name
    // whom the login action was called on
    if ($pagename != $LoginPage) {
      if (isset($_SERVER['HTTP_REFERER']) && $_SERVER['HTTP_REFERER'] != "") {
	$_SESSION['previous_page'] = $_SERVER['HTTP_REFERER'];
      } else {
	$_SESSION['previous_page'] = $pagename;
      }
    }
    
    // Make sure both that a login page is defined and that we
    // are not currently at the login page. If no default abilities
    // are defined then the system would keep redirecting to the
    // login page since it is handled by PmWiki. Therefore if the
    // current page is login page when a login action is requested, we
    // just print out the backup standalone page.
    if (PageExists($LoginPage)) {
      // Redirect to the login page to display username password form,
      // but if the Wiki is totally restricted and $LoginPage is still
      // defined then print a simple standalone page, because if we
      // did redirect when the GuestUser had no read ability then we
      // would keep on redirect so long as the browser lets us!
      if($UserInfoObj->UserHasAbility($GuestUsername, "read")) { 
	Redirect($LoginPage);
      } else {
	PrintStandalonePage( GetUserLoginForm(false) );
      }
    } else {
      PrintEmbeddedPage( $pagename, GetUserLoginForm(false) );
    }
  }
}

function HandleLogout($pagename) {
  global $UserInstanceVars, $RedirectToPrevious, $DefaultRedirectionPage;

  // Clear UserAuth instance variables
  $UserInstanceVars->ClearInstanceVariables();

  // PmWiki BasicAuth password storeage array
  unset( $_SESSION['authpw'] );

  // Use referer for redirection if it is available
  if ( $RedirectToPrevious && 
       isset($_SERVER['HTTP_REFERER']) && $_SERVER['HTTP_REFERER'] != "" ) {

    RedirectURL($_SERVER['HTTP_REFERER']);

  } else {
    if(isset($DefaultRedirectionPage)) {
      Redirect($DefaultRedirectionPage);
    } else {
      Redirect($pagename);
    }
  }
}

function RedirectURL($url) {
  global $RedirectDelay;
  clearstatcache();

  header("Location: $url");
  header("Content-type: text/html");
  print("<html><head>
    <meta http-equiv='Refresh' Content='$RedirectDelay; URL=$url'>
    <title>Redirect</title></head><body></body></html>");
  exit;
}

?>
